"""
Copyright 2021 Demetry Pascal (forked from Ryan (Mohammad) Solgi)

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to use, 
copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the 
Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
SOFTWARE.
"""


from typing import Callable, List, Tuple, Optional, Dict, Any, Union, Sequence, Set

import collections
import warnings
import operator


import sys
import time
import random
import math

import numpy as np
from func_timeout import func_timeout, FunctionTimedOut

from OppOpPopInit import init_population, SampleInitializers, OppositionOperators, set_seed

#region INTERNAL IMPORTS

from .aliases import array1D, array2D

from .classes import AlgorithmParams, Generation, MiddleCallbackData, GAResult, GenerationConvertible

from .initializer import Population_initializer
from .plotting_tools import plot_pop_scores, plot_several_lines

from .utils import can_be_prob, is_numpy, is_current_gen_number, fast_min, random_indexes_pair

from .callbacks import MiddleCallbackFunc, CallbackFunc

#endregion


class geneticalgorithm2:
    
    """
    Genetic Algorithm (Elitist version) for Python
    
    An implementation of elitist genetic algorithm for solving problems with
    continuous, integers, or mixed variables.
    
    repo path https://github.com/PasaOpasen/geneticalgorithm2
    """
    
    default_params = AlgorithmParams()
    PROGRESS_BAR_LEN = 20

    @property
    def output_dict(self):
        warnings.warn(
            "'output_dict' is deprecated and will be removed at version 7 \n use 'result' instead"
        )
        return self.result

    @property
    def needs_mutation(self) -> bool:
        return self.prob_mut > 0 or self.prob_mut_discrete > 0

    #region INIT

    def __init__(
        self,
        function: Callable[[array1D], float],

        dimension: int,
        variable_type: Union[str, Sequence[str]] = 'bool',
        variable_boundaries: Optional[Union[array2D, Sequence[Tuple[float, float]]]] = None,

        variable_type_mixed=None,

        function_timeout: float = 10,
        algorithm_parameters: Union[AlgorithmParams, Dict[str, Any]] = default_params
    ):
        """
        @param function <Callable[[np.ndarray], float]> - the given objective function to be minimized
        NOTE: This implementation minimizes the given objective function.
        (For maximization multiply function by a negative sign: the absolute
        value of the output would be the actual objective function)

        @param dimension <integer> - the number of decision variables

        @param variable_type <string> - 'bool' if all variables are Boolean;
        'int' if all variables are integer; and 'real' if all variables are
        real value or continuous. For mixed types use sequence of string of type for each variable

        @param variable_boundaries <Optional[Union[np.ndarray, Sequence[Tuple[float, float]]]]> - Default None; leave it
        None if variable_type is 'bool'; otherwise provide an array of tuples
        of length two as boundaries for each variable;
        the length of the array must be equal dimension. For example,
        np.array([0,100],[0,200]) determines lower boundary 0 and upper boundary 100 for first
        and upper boundary 200 for second variable where dimension is 2.

        @param variable_type_mixed -- deprecated

        @param function_timeout <float> - if the given function does not provide
        output before function_timeout (unit is seconds) the algorithm raise error.
        For example, when there is an infinite loop in the given function.

        @param algorithm_parameters <Union[AlgorithmParams, Dict[str, Any]]>:
            @ max_num_iteration <int> - stoping criteria of the genetic algorithm (GA)
            @ population_size <int>
            @ mutation_probability <float in [0,1]>
            @ elit_ratio <float in [0,1]>
            @ crossover_probability <float in [0,1]>
            @ parents_portion <float in [0,1]>
            @ crossover_type <string/function> - Default is 'uniform'; 'one_point' or 'two_point' (not only) are other options
            @ mutation_type <string/function> - Default is 'uniform_by_x'; see GitHub to check other options
            @ mutation_discrete_type <string/function> - mutation type for discrete variables
            @ selection_type <string/function> - Default is 'roulette'; see GitHub to check other options
            @ max_iteration_without_improv <int> - maximum number of successive iterations without improvement. If None it is ineffective


        for more details and examples of implementation please visit:
            https://github.com/PasaOpasen/geneticalgorithm2
  
        """


        # all default fields

        self.param: AlgorithmParams = None
        # self.crossover: Callable[[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]] = None
        # self.real_mutation: Callable[[float, float, float], float] = None
        # self.discrete_mutation: Callable[[int, int, int], int] = None
        # self.selection: Callable[[np.ndarray, int], np.ndarray] = None

        self.f: Callable[[array1D], float] = None
        self.funtimeout: float = None
        self.set_function: Callable[[np.ndarray], np.ndarray] = None

        # self.dim: int = None
        self.var_bounds: List[Tuple[Union[int, float], Union[int, float]]] = None
        # self.indexes_int: np.ndarray = None
        # self.indexes_float: np.ndarray = None

        self.checked_reports: List[Tuple[str, Callable[[array1D], None]]] = None

        self.population_size: int = None
        self.progress_stream = None


        # input algorithm's parameters

        assert type(algorithm_parameters) in (dict, AlgorithmParams), "algorithm_parameters must be dict or AlgorithmParams object"
        if type(algorithm_parameters) != AlgorithmParams:
            algorithm_parameters = AlgorithmParams.from_dict(algorithm_parameters)

        algorithm_parameters.check_if_valid()
        self.param = algorithm_parameters
        self.crossover, self.real_mutation, self.discrete_mutation, self.selection = algorithm_parameters.get_CMS_funcs()


        # dimension and area bounds
        self.dim = int(dimension)
        assert self.dim > 0, f"dimension count must be int and >0, gotten {dimension}"


        if variable_type_mixed is not None:
            warnings.warn(
                f"argument variable_type_mixed is deprecated and will be removed at version 7\n use variable_type={tuple(variable_type_mixed)} instead"
            )
            variable_type = variable_type_mixed
        self._set_types_indexes(variable_type)  # types indexes
        self._set_var_boundaries(variable_type, variable_boundaries)  # input variables' boundaries


        # fix mutation probs

        assert can_be_prob(self.param.mutation_probability)
        self.prob_mut = self.param.mutation_probability
        assert self.param.mutation_discrete_probability is None or can_be_prob(self.param.mutation_discrete_probability)
        self.prob_mut_discrete = self.param.mutation_discrete_probability or self.prob_mut

        if self.param.crossover_probability is not None:
            warnings.warn(
                f"crossover_probability is deprecated and will be removed in version 7. Reason: it's old and has no sense"
            )

        #############################################################
        # input function
        assert (callable(function)), "function must be callable!"
        self.f = function
        self.funtimeout = float(function_timeout)

        #############################################################
        
        self.population_size = int(self.param.population_size)
        self._set_parents_count(self.param.parents_portion)
        self._set_elit_count(self.population_size, self.param.elit_ratio)
        assert(self.parents_count >= self.elit_count), f"\n number of parents ({self.parents_count}) must be greater than number of elits ({self.elit_count})"

        self._set_max_iterations()

        self._set_report()

        # specify this function to speed up or change default behaviour
        self.fill_children: Optional[Callable[[array2D, int], None]] = None
        """custom function which adds children for population POP where POP[:parents_count] are parents lines and next lines are for children"""

    def _set_types_indexes(self, variable_type: Union[str, Sequence[str]]):

        indexes = np.arange(self.dim)
        self.indexes_int = np.array([])
        self.indexes_float = np.array([])

        VALID_STRINGS = ('bool', 'int', 'real')
        assert_message = f"\n variable_type must be 'bool', 'int', 'real' or a sequence with 'int' and 'real', got {variable_type}"

        if type(variable_type) == str:
            assert (variable_type in VALID_STRINGS), assert_message
            if variable_type == 'real':
                self.indexes_float = indexes
            else:
                self.indexes_int = indexes

        else:  # sequence case

            assert (len(variable_type) == self.dim), f"\n variable_type must have a length equal dimension. Should be {self.dim}, got {len(variable_type)}"
            assert 'bool' not in variable_type, "don't use 'bool' if variable_type is a sequence, for 'boolean' case use 'int' and specify boundary as (0,1)"
            assert all(v in VALID_STRINGS for v in variable_type), assert_message

            vartypes = np.array(variable_type)
            self.indexes_int = indexes[vartypes == 'int']
            self.indexes_float = indexes[vartypes == 'real']

    def _set_var_boundaries(
        self,
        variable_type: Union[str, Sequence[str]],
        variable_boundaries
    ):
        if isinstance(variable_type, str) and variable_type == 'bool':
            self.var_bounds = [(0, 1)] * self.dim
        else:

            if is_numpy(variable_boundaries):
                assert variable_boundaries.shape == (self.dim, 2), f"\n if variable_boundaries is numpy array, it must be with shape (dim, 2)"
            else:
                assert len(variable_boundaries) == self.dim and all((len(t) == 2 for t in variable_boundaries)), "\n if variable_boundaries is sequence, it must be with len dim and boundary for each variable must be a tuple of length two"

            for i in variable_boundaries:
                assert(i[0] <= i[1]), "\n lower_boundaries must be smaller than upper_boundaries [lower,upper]"

            self.var_bounds = [(i[0], i[1]) for i in variable_boundaries]


    def _set_parents_count(self, parents_portion: float):

        self.parents_count = int(parents_portion * self.population_size)
        assert self.parents_count <= self.population_size and self.parents_count > 1, f'parents count {self.parents_count} cannot be less than population size {self.population_size}'
        trl = self.population_size - self.parents_count
        if trl % 2 != 0:
            self.parents_count += 1

    def _set_elit_count(self, pop_size: int, elit_ratio: Union[float, int]):

        assert elit_ratio >= 0
        self.elit_count = elit_ratio if type(elit_ratio) == int else math.ceil(pop_size*elit_ratio)

    def _set_max_iterations(self):

        if self.param.max_num_iteration is None:
            iterate = 0
            for i in range(0, self.dim):
                bound_min, bound_max = self.var_bounds[i]
                var_space = bound_max - bound_min
                if i in self.indexes_int:
                    iterate += var_space * self.dim * (100 / self.population_size)
                else:
                    iterate += var_space * 50 * (100 / self.population_size)
            iterate = int(iterate)
            if (iterate * self.population_size) > 10000000:
                iterate = 10000000 / self.population_size

            self.max_iterations = fast_min(iterate, 8000)
        else:
            assert self.param.max_num_iteration > 0
            self.max_iterations = math.ceil(self.param.max_num_iteration)

        max_it = self.param.max_iteration_without_improv
        if max_it is None:
            self.max_stagnations = self.max_iterations + 1
        else:
            self.max_stagnations = math.ceil(max_it)

    #endregion

    #region REPORT

    def _set_report(self):
        """
        creates default report checker
        """
        self.checked_reports = [
            # item 0 cuz scores will be sorted and min item is items[0]
            ('report', operator.itemgetter(0))
        ]

    def _clear_report(self):
        """
        removes all report objects
        """
        fields = [f for f in vars(self).keys() if f.startswith('report')]
        for attr in fields:
            delattr(self, attr)

    def _init_report(self):
        """
        makes empty report fields
        """
        for name, _ in self.checked_reports:
            setattr(self, name, [])

    def _update_report(self, scores: array1D):
        """
        append report value to the end of field
        """
        for name, func in self.checked_reports:
            getattr(self, name).append(
                func(scores)
            )

    #endregion

    #region RUN METHODS

    def __progress(self, count: int, total: int, status: str = ''):

        part = count / total

        filled_len = round(geneticalgorithm2.PROGRESS_BAR_LEN * part)
        percents = round(100.0 * part, 1)
        bar = '|' * filled_len + '_' * (geneticalgorithm2.PROGRESS_BAR_LEN - filled_len)

        self.progress_stream.write('\r%s %s%s %s' % (bar, percents, '%', status))
        self.progress_stream.flush()

    def __str__(self):
        return f"Genetic algorithm object with parameters {self.param}"

    def __repr__(self):
        return self.__str__()

    def _simulate(self, sample: array1D):

        obj = None
        eval_time = time.time()
        try:
            obj = func_timeout(
                self.funtimeout,
                lambda: self.f(sample)
            )
        except FunctionTimedOut:
            print("given function is not applicable")
        eval_time = time.time() - eval_time

        assert obj is not None, f"the given function was running like {eval_time} seconds and does not provide any output"

        tp = type(obj)
        assert (
            tp in (int, float) or np.issubdtype(tp, np.floating) or np.issubdtype(tp, np.integer)
        ), f"Minimized function should return a number, but got '{obj}' object with type {tp}"

        return obj, eval_time

    def _set_mutation_indexes(self, mutation_indexes: Optional[Sequence[int]]):

        if mutation_indexes is None:
            self.indexes_float_mut = self.indexes_float
            self.indexes_int_mut = self.indexes_int
        else:
            tmp_indexes = set(mutation_indexes)
            self.indexes_int_mut = np.array(list(set(self.indexes_int).intersection(tmp_indexes)))
            self.indexes_float_mut = np.array(list(set(self.indexes_float).intersection(tmp_indexes)))

            if self.indexes_float_mut.size == 0 and self.indexes_int_mut.size == 0:
                warnings.warn(f"No mutation dimensions!!! Check ur mutation indexes!!")

    #@profile
    def run(
        self,
        no_plot: bool = False,
        disable_printing: bool = False,
        progress_bar_stream: Optional[str] = 'stdout',

        # deprecated
        disable_progress_bar: bool = False,

        set_function: Optional[Callable[[array2D], array1D]] = None,
        apply_function_to_parents: bool = False,
        start_generation: GenerationConvertible = Generation(),
        studEA: bool = False,
        mutation_indexes: Optional[Union[Sequence[int], Set[int]]] = None,

        init_creator: Optional[Callable[[], array1D]] = None,
        init_oppositors: Optional[Sequence[Callable[[array1D], array1D]]] = None,

        duplicates_oppositor: Optional[Callable[[array1D], array1D]] = None,
        remove_duplicates_generation_step: Optional[int] = None,

        revolution_oppositor: Optional[Callable[[array1D], array1D]] = None,
        revolution_after_stagnation_step: Optional[int] = None,
        revolution_part: float = 0.3,

        population_initializer: Tuple[int, Callable[[array2D, array1D], Tuple[array2D, array1D]]] = Population_initializer(select_best_of=1, local_optimization_step='never', local_optimizer=None),

        stop_when_reached: Optional[float] = None,
        callbacks: Optional[Sequence[CallbackFunc]] = None,
        middle_callbacks: Optional[Sequence[MiddleCallbackFunc]] = None, #+
        time_limit_secs: Optional[float] = None,
        save_last_generation_as: Optional[str] = None,
        seed: Optional[int] = None
    ):
        """
        runs optimization process

        @param no_plot: do not plot results using matplotlib by default

        @param disable_printing: do not print log info of optimization process
        
        @param progress_bar_stream: 'stdout', 'stderr' or None to disable progress bar

        @param disable_progress_bar:
                
        @param set_function : 2D-array -> 1D-array function, which applyes to matrix of population (size (samples, dimention))
        to estimate their values
        
        @param apply_function_to_parents: apply function to parents from previous generation (if it's needed)
                                                                                                         
        @param start_generation: Generation object or a dictionary with structure {'variables':2D-array of samples, 'scores': function values on samples} or path to .npz file (str) with saved generation; if 'scores' value is None the scores will be compute

        @param studEA: using stud EA strategy (crossover with best object always)
        
        @param mutation_indexes: indexes of dimensions where mutation can be performed (all dimensions by default)

        @param init_creator: the function creates population samples. By default -- random uniform for real variables and random uniform for int
        @param init_oppositors: the list of oppositors creates oppositions for base population. No by default
        @param duplicates_oppositor: oppositor for applying after duplicates removing. By default -- using just random initializer from creator
        @param remove_duplicates_generation_step: None/int, step for removing duplicates (have a sense with discrete tasks). No by default
        @param revolution_oppositor: oppositor for revolution time. No by default
        @param revolution_after_stagnation_step: create revolution after this generations of stagnation. No by default
        @param revolution_part: float, the part of generation to being oppose. By default is 0.3

        @param population_initializer: object for actions at population initialization step to create better start population. See doc

        @param stop_when_reached: stop searching after reaching this value (it can be potential minimum or something else)

        @param callbacks: sequence of callback functions with structure: (generation_number, report_list, last_population, last_scores) -> do some action

        @param middle_callbacks: sequence of functions made MiddleCallbacks class

        @param time_limit_secs: limit time of working (in seconds)

        @param save_last_generation_as: path to .npz file for saving last_generation as numpy dictionary like {'population': 2D-array, 'scores': 1D-array}, None if doesn't need to save in file

        @param seed: random seed (None if doesn't matter)
        """

        if disable_progress_bar:
            warnings.warn(
                f"disable_progress_bar is deprecated and will be removed in version 7, use probress_bar_stream=None to disable progress bar"
            )
            progress_bar_stream = None

        enable_printing: bool = not disable_printing

        start_generation = Generation.from_object(self.dim, start_generation)

        assert is_current_gen_number(revolution_after_stagnation_step), "must be None or int and >0"
        assert is_current_gen_number(remove_duplicates_generation_step), "must be None or int and >0"
        assert can_be_prob(revolution_part), f"revolution_part must be in [0,1], not {revolution_part}"
        assert (stop_when_reached is None or type(stop_when_reached) in (int, float))
        assert (isinstance(callbacks, collections.abc.Sequence) or callbacks is None), "callbacks should be list of callbacks functions"
        assert (isinstance(middle_callbacks, collections.abc.Sequence) or middle_callbacks is None), "middle_callbacks should be list of MiddleCallbacks functions"
        assert (time_limit_secs is None or time_limit_secs > 0), 'time_limit_secs must be None of number > 0'

        self._set_mutation_indexes(mutation_indexes)
        set_seed(seed)

        # randomstate = np.random.default_rng(random.randint(0, 1000) if seed is None else seed)
        # self.randomstate = randomstate

        # using bool flag is faster than using empty function with generated string arguments
        SHOW_PROGRESS = progress_bar_stream is not None
        if SHOW_PROGRESS:

            show_progress = lambda t, t2, s: self.__progress(t, t2, status=s)

            if progress_bar_stream == 'stdout':
                self.progress_stream = sys.stdout
            elif progress_bar_stream == 'stderr':
                self.progress_stream = sys.stderr
            else:
                raise Exception(
                    f"wrong value {progress_bar_stream} of progress_bar_stream, must be 'stdout'/'stderr'/None"
                )
        else:
            show_progress = None

        stop_by_val = (lambda best_f: False) if stop_when_reached is None else (lambda best_f: best_f <= stop_when_reached)

        t: int = 0
        count_stagnation: int = 0
        pop: array2D = None
        scores: array1D = None

        #
        #
        # callbacks part
        #
        #

        def get_data():
            """
            returns all important data about model
            """
            return MiddleCallbackData(
                last_generation=Generation(pop, scores),
                current_generation=t,
                report_list=self.report,

                mutation_prob=self.prob_mut,
                mutation_discrete_prob=self.prob_mut_discrete,
                mutation=self.real_mutation,
                mutation_discrete=self.discrete_mutation,
                crossover=self.crossover,
                selection=self.selection,

                current_stagnation=count_stagnation,
                max_stagnation=self.max_stagnations,

                parents_portion=self.param.parents_portion,
                elit_ratio=self.param.elit_ratio,

                set_function=self.set_function,

                reason_to_stop=reason_to_stop
            )

        def set_data(data: MiddleCallbackData):
            """
            sets data to model
            """
            nonlocal pop, scores, count_stagnation, reason_to_stop

            pop, scores = data.last_generation.variables, data.last_generation.scores
            self.population_size = pop.shape[0]

            self.param.parents_portion = data.parents_portion
            self._set_parents_count(data.parents_portion)

            self.param.elit_ratio = data.elit_ratio
            self._set_elit_count(self.population_size, data.elit_ratio)

            self.prob_mut = data.mutation_prob
            self.prob_mut_discrete = data.mutation_discrete_prob
            self.real_mutation = data.mutation
            self.discrete_mutation = data.mutation_discrete
            self.crossover = data.crossover
            self.selection = data.selection

            count_stagnation = data.current_stagnation
            reason_to_stop = data.reason_to_stop
            self.max_stagnations = data.max_stagnation

            self.set_function = data.set_function

        if not callbacks:
            total_callback = lambda g, r, lp, ls: None
        else:
            def total_callback(
                generation_number: int,
                report_list: List[float],
                last_population: array2D,
                last_scores: array1D
            ):
                for cb in callbacks:
                    cb(generation_number, report_list, last_population, last_scores)

        if not middle_callbacks:
            total_middle_callback = lambda: None
        else:
            def total_middle_callback():
                """
                applies callbacks and sets new data if there is a sence
                """
                data = get_data()
                flag = False
                for cb in middle_callbacks:
                    data, has_sense = cb(data)
                    if has_sense:
                        flag = True
                if flag:
                    set_data(data)  # update global date if there was real callback step

        ############################################################

        start_time = time.time()
        time_is_done = (lambda: False) if time_limit_secs is None else (lambda: int(time.time() - start_time) >= time_limit_secs)

        ############################################################# 
        # Initial Population
        self.set_function = set_function or geneticalgorithm2.default_set_function(self.f)

        pop_coef, initializer_func = population_initializer
        
        # population creator by random or with oppositions
        if init_creator is None:
            # just uniform random
            self.creator = SampleInitializers.Combined(
                minimums=[v[0] for v in self.var_bounds],
                maximums=[v[1] for v in self.var_bounds],
                indexes=(
                    self.indexes_int,
                    self.indexes_float
                ),
                creator_initializers=(
                    SampleInitializers.RandomInteger,
                    SampleInitializers.Uniform
                )
            )
        else:
            assert callable(init_creator)
            self.creator = init_creator

        self.init_oppositors = init_oppositors
        self.dup_oppositor = duplicates_oppositor
        self.revolution_oppositor = revolution_oppositor

        # event for removing duplicates
        if remove_duplicates_generation_step is None:
            def remover(pop: array2D, scores: array1D, gen: int) -> Tuple[
                array2D,
                array1D
            ]:
                return pop, scores
        else:
            
            def without_dup(pop: array2D, scores: array1D):  # returns population without dups
                _, index_of_dups = np.unique(pop, axis=0, return_index=True)
                return pop[index_of_dups], scores[index_of_dups], scores.size - index_of_dups.size
            
            if self.dup_oppositor is None:  # if there is no dup_oppositor, use random creator
                def remover(pop: array2D, scores: array1D, gen: int) -> Tuple[
                    array2D,
                    array1D
                ]:
                    if gen % remove_duplicates_generation_step != 0:
                        return pop, scores

                    pp, sc, count_to_create = without_dup(pop, scores)  # pop without dups
                    
                    if count_to_create == 0:
                        if SHOW_PROGRESS:
                            show_progress(t, self.max_iterations,
                                      f"GA is running...{t} gen from {self.max_iterations}. No dups!")
                        return pop, scores

                    pp2 = SampleInitializers.CreateSamples(self.creator, count_to_create)  # new pop elements
                    pp2_scores = self.set_function(pp2)  # new elements values

                    if SHOW_PROGRESS:
                        show_progress(t, self.max_iterations,
                                      f"GA is running...{t} gen from {self.max_iterations}. Kill dups!")
                    
                    new_pop = np.vstack((pp, pp2))
                    new_scores = np.concatenate((sc, pp2_scores))

                    args_to_sort = new_scores.argsort()
                    return new_pop[args_to_sort], new_scores[args_to_sort]

            else:  # using oppositors
                assert callable(self.dup_oppositor)

                def remover(pop: np.ndarray, scores: np.ndarray, gen: int) -> Tuple[
                    np.ndarray,
                    np.ndarray
                ]:
                    if gen % remove_duplicates_generation_step != 0:
                        return pop, scores

                    pp, sc, count_to_create = without_dup(pop, scores)  # pop without dups

                    if count_to_create == 0:
                        if SHOW_PROGRESS:
                            show_progress(t, self.max_iterations,
                                          f"GA is running...{t} gen from {self.max_iterations}. No dups!")
                        return pop, scores

                    if count_to_create > sc.size:
                        raise Exception(f"Too many duplicates at generation {gen} ({count_to_create} > {sc.size}), cannot oppose")

                    # oppose count_to_create worse elements
                    pp2 = OppositionOperators.Reflect(pp[-count_to_create:], self.dup_oppositor)  # new pop elements
                    pp2_scores = self.set_function(pp2)  # new elements values

                    if SHOW_PROGRESS:
                        show_progress(t, self.max_iterations,
                                      f"GA is running...{t} gen from {self.max_iterations}. Kill dups!")

                    new_pop = np.vstack((pp, pp2))
                    new_scores = np.concatenate((sc, pp2_scores))

                    args_to_sort = new_scores.argsort()
                    return new_pop[args_to_sort], new_scores[args_to_sort]

        # event for revolution
        if revolution_after_stagnation_step is None:
            def revolution(pop: array2D, scores: array1D, stagnation_count: int) -> Tuple[
                array2D,
                array1D
            ]:
                return pop, scores
        else:
            if revolution_oppositor is None:
                raise Exception(
                    f"How can I make revolution each {revolution_after_stagnation_step} stagnation steps if revolution_oppositor is None (not defined)?"
                )
            assert callable(revolution_oppositor)
            
            def revolution(pop: array2D, scores: array1D, stagnation_count: int) -> Tuple[
                array2D,
                array1D
            ]:
                if stagnation_count < revolution_after_stagnation_step:
                    return pop, scores
                part = int(pop.shape[0]*revolution_part)
                
                pp2 = OppositionOperators.Reflect(pop[-part:], self.revolution_oppositor)
                pp2_scores = self.set_function(pp2)

                combined = np.vstack((pop, pp2))
                combined_scores = np.concatenate((scores, pp2_scores))
                args = combined_scores.argsort()

                if SHOW_PROGRESS:
                    show_progress(t, self.max_iterations,
                                  f"GA is running...{t} gen from {self.max_iterations}. Revolution!")

                args = args[:scores.size]
                return combined[args], combined_scores[args]

        #
        #
        #  START ALGORITHM LOGIC
        #
        #

        # Report
        self._clear_report()  # clear old report objects
        self._init_report()

        # initialization of pop
        
        if start_generation.variables is None:

            real_pop_size = self.population_size * pop_coef

            # pop = np.empty((real_pop_size, self.dim))
            scores = np.empty(real_pop_size)
            
            pop = init_population(
                samples_count=real_pop_size,
                creator=self.creator,
                oppositors=self.init_oppositors
            )
            
            time_counter = 0

            for p in range(0, real_pop_size):
                # simulation returns exception or func value -- check the time of evaluating
                value, eval_time = self._simulate(pop[p])
                scores[p] = value
                time_counter += eval_time
            
            if enable_printing:
                print(
                    f"\nAverage time of function evaluating (secs): {time_counter/real_pop_size} (total = {time_counter})\n"
                )
                
        else:

            self.population_size = start_generation.variables.shape[0]
            self._set_elit_count(self.population_size, self.param.elit_ratio)
            self._set_parents_count(self.param.parents_portion)

            pop = start_generation.variables

            if start_generation.scores is None:

                _time = time.time()
                scores = self.set_function(pop)
                _time = time.time() - _time

                if enable_printing:
                    print(
                        f'\nFirst scores are made from gotten variables (by {_time} secs, about {_time/scores.size} for each creature)\n'
                    )
            else:
                scores = start_generation.scores
                if enable_printing:
                    print(f"\nFirst scores are from gotten population\n")


        # Initialization by select bests and local_descent
        
        pop, scores = initializer_func(pop, scores)

        # first sort
        args_to_sort = scores.argsort()
        pop = pop[args_to_sort]
        scores = scores[args_to_sort]
        self._update_report(scores)

        self.population_size = scores.size
        self.best_function = scores[0]

        if enable_printing:
            print(
                f"Best score before optimization: {self.best_function}"
            )

        t: int = 1
        count_stagnation: int = 0
        """iterations without progress"""
        reason_to_stop: Optional[str] = None

        # gets indexes of 2 parents to crossover
        if studEA:
            get_parents_inds = lambda: (0, random.randrange(1, self.parents_count))
        else:
            get_parents_inds = lambda: random_indexes_pair(self.parents_count)

        #  while no reason to stop
        while True:

            if count_stagnation > self.max_stagnations:
                reason_to_stop = f"limit of fails: {count_stagnation}"
            elif t == self.max_iterations:
                reason_to_stop = f'limit of iterations: {t}'
            elif stop_by_val(self.best_function):
                reason_to_stop = f"stop value reached: {self.best_function} <= {stop_when_reached}"
            elif time_is_done():
                reason_to_stop = f'time is done: {time.time() - start_time} >= {time_limit_secs}'

            if reason_to_stop is not None:
                if SHOW_PROGRESS:
                    show_progress(t, self.max_iterations, f"GA is running... STOP! {reason_to_stop}")
                break

            if scores[0] < self.best_function:  # if there is progress
                count_stagnation = 0
                self.best_function = scores[0]
            else:
                count_stagnation += 1

            if SHOW_PROGRESS:
                show_progress(
                    t,
                    self.max_iterations,
                    f"GA is running...{t} gen from {self.max_iterations}...best value = {self.best_function}"
                )

            # Select parents
            
            par: array2D = np.empty((self.parents_count, self.dim))
            """samples chosen to create new samples"""
            par_scores: array1D = np.empty(self.parents_count)

            elit_slice = slice(None, self.elit_count)
            """elit parents"""

            # copy needs because the generation will be removed after parents selection
            par[elit_slice] = pop[elit_slice].copy()
            par_scores[elit_slice] = scores[elit_slice].copy()

            new_par_inds = self.selection(scores[self.elit_count:], self.parents_count - self.elit_count).astype(np.int16) + self.elit_count
            """non-elit parents indexes"""
            #new_par_inds = self.selection(scores, self.parents_count - self.elit_count).astype(np.int16)
            par_slice = slice(self.elit_count, self.parents_count)
            par[par_slice] = pop[new_par_inds].copy()
            par_scores[par_slice] = scores[new_par_inds].copy()

            pop = np.empty((self.population_size, self.dim))
            """new generation"""
            scores = np.empty(self.population_size)
            """new generation scores"""

            parents_slice = slice(None, self.parents_count)
            pop[parents_slice] = par
            scores[parents_slice] = par_scores

            if self.fill_children is None:  # default fill children behaviour
                DO_MUTATION = self.needs_mutation
                for k in range(self.parents_count, self.population_size, 2):

                    r1, r2 = get_parents_inds()

                    pvar1 = pop[r1]  # equal to par[r1], but better for cache
                    pvar2 = pop[r2]

                    ch1, ch2 = self.crossover(pvar1, pvar2)

                    if DO_MUTATION:
                        ch1 = self.mut(ch1)
                        ch2 = self.mut_middle(ch2, pvar1, pvar2)

                    pop[k] = ch1
                    pop[k+1] = ch2
            else:  # custom behaviour
                self.fill_children(pop, self.parents_count)

            if apply_function_to_parents:
                scores = self.set_function(pop)
            else:
                scores[self.parents_count:] = self.set_function(pop[self.parents_count:])
            
            # remove duplicates
            pop, scores = remover(pop, scores, t)
            # revolution
            pop, scores = revolution(pop, scores, count_stagnation)

            # make population better
            args_to_sort = scores.argsort()
            pop = pop[args_to_sort]
            scores = scores[args_to_sort]
            self._update_report(scores)

            # callback it
            total_callback(t, self.report, pop, scores)
            total_middle_callback()

            t += 1

        self.best_function = fast_min(scores[0], self.best_function)

        last_generation = Generation(pop, scores)
        self.result = GAResult(last_generation)

        if save_last_generation_as is not None:
            last_generation.save(save_last_generation_as)

        if enable_printing:
            show = ' '*200
            sys.stdout.write(f'\r{show}\n')
            sys.stdout.write(f'\r The best found solution:\n {pop[0]}')
            sys.stdout.write(f'\n\n Objective function:\n {self.best_function}\n')
            sys.stdout.write(f'\n Used generations: {len(self.report)}')
            sys.stdout.write(f'\n Used time: {time.time() - start_time:.3g} seconds\n')
            sys.stdout.flush() 
        
        if not no_plot:
            self.plot_results()

        if enable_printing:
            if reason_to_stop is not None and 'iterations' not in reason_to_stop:
                sys.stdout.write(
                    f'\nWarning: GA is terminated because of {reason_to_stop}'
                )

        return self.result

    #endregion

    #region PLOTTING

    def plot_results(
        self,
        title: str = 'Genetic Algorithm',
        save_as: Optional[str] = None,
        dpi: int = 200,
        main_color: str = 'blue'
     ):
        """
        Simple plot of self.report (if not empty)
        """
        if len(self.report) == 0:
            sys.stdout.write("No results to plot!\n")
            return

        plot_several_lines(
            lines=[self.report],
            colors=[main_color],
            labels=['best of generation'],
            linewidths=[2],
            title=title,
            xlabel='Generation',
            ylabel='Minimized function',
            save_as=save_as,
            dpi=dpi
        )

    def plot_generation_scores(self, title: str = 'Last generation scores', save_as: Optional[str] = None):
        """
        Plots barplot of scores of last population
        """

        if not hasattr(self, 'result'):
            raise Exception("There is no 'result' field into ga object! Before plotting generation u need to run seaching process")

        plot_pop_scores(self.result.last_generation.scores, title, save_as)

    #endregion

    #region MUTATION
    def mut(self, x: array1D):
        """
        just mutation
        """

        for i in self.indexes_int_mut:
            if random.random() < self.prob_mut_discrete:
                x[i] = self.discrete_mutation(x[i], *self.var_bounds[i])

        for i in self.indexes_float_mut:                
            if random.random() < self.prob_mut:
                x[i] = self.real_mutation(x[i], *self.var_bounds[i])
            
        return x

    def mut_middle(self, x: array1D, p1: array1D, p2: array1D):
        """
        mutation oriented on parents
        """
        for i in self.indexes_int_mut:

            if random.random() < self.prob_mut_discrete:
                v1, v2 = p1[i], p2[i]
                if v1 < v2:
                    x[i] = random.randint(v1, v2)
                elif v1 > v2:
                    x[i] = random.randint(v2, v1)
                else:
                    x[i] = random.randint(*self.var_bounds[i])
                        
        for i in self.indexes_float_mut:                
            if random.random() < self.prob_mut:
                v1, v2 = p1[i], p2[i]
                if v1 != v2:
                    x[i] = random.uniform(v1, v2)
                else:
                    x[i] = random.uniform(*self.var_bounds[i])
        return x

    #endregion


    #region Set functions


    @staticmethod
    def default_set_function(function_for_set: Callable[[array1D], float]):
        """
        simple function for creating set_function 
        function_for_set just applyes to each row of population
        """
        def func(matrix: array2D):
            return np.array(
                [function_for_set(row) for row in matrix]
            )
        return func

    @staticmethod
    def vectorized_set_function(function_for_set: Callable[[array1D], float]):
        """
        works like default, but faster for big populations and slower for little
        function_for_set just applyes to each row of population
        """

        func = np.vectorize(function_for_set, signature='(n)->()')
        return func

    @staticmethod
    def set_function_multiprocess(function_for_set: Callable[[array1D], float], n_jobs: int = -1):
        """
        like function_for_set but uses joblib with n_jobs (-1 goes to count of available processors)
        """
        from joblib import Parallel, delayed

        def func(matrix: array2D):
            result = Parallel(n_jobs=n_jobs)(delayed(function_for_set)(matrix[i]) for i in range(matrix.shape[0]))
            return np.array(result)
        return func
            
    #endregion













            
            



from typing import Callable, Optional, Tuple, Literal

import numpy as np

from .aliases import TypeAlias, array1D, array2D


LOCAL_OPT_STEPS: TypeAlias = Literal['before_select', 'after_select', 'never']


def Population_initializer(
    select_best_of: int = 4,
    local_optimization_step: LOCAL_OPT_STEPS = 'never',
    local_optimizer: Optional[
        Callable[
            [array1D, float],
            Tuple[array1D, float]
        ]
    ] = None
) -> Tuple[int, Callable[[array2D, array1D], Tuple[array2D, array1D]]]:
    """
    select_best_of (int) -- select 1/select_best_of best part of start population. For example, for select_best_of = 4 and population_size = N will be selected N best objects from 5N generated objects (if start_generation = None dictionary). If start_generation is not None dictionary, it will be selected best size(start_generation)/N  objects

    local_optimization_step (str) -- when should we do local optimization? Available values:
    
    * 'never' -- don't do local optimization
    * 'before_select' -- before selection best N objects (example: do local optimization for 5N objects and select N best results)
    * 'after_select' -- do local optimization on best selected N objects

    local_optimizer (function) -- local optimization function like:
        def loc_opt(object_as_array, current_score):
            # some code
            return better_object_as_array, better_score
    """
    
    assert (select_best_of > 0 and type(select_best_of) == int), "select_best_of argument should be integer and more than 0"

    assert (local_optimization_step in LOCAL_OPT_STEPS.__args__), f"local_optimization_step should be in {LOCAL_OPT_STEPS.__args__}, but got {local_optimization_step}"

    if local_optimizer is None and local_optimization_step in LOCAL_OPT_STEPS.__args__[:2]:
        raise Exception(f"for local_optimization_step from {LOCAL_OPT_STEPS.__args__[:2]} local_optimizer function mustn't be None")


    def Select_best(population: array2D, scores: array1D) -> Tuple[array2D, array1D]:
        args = np.argsort(scores)
        args = args[:round(args.size/select_best_of)]
        return population[args], scores[args]

    def Local_opt(population: array2D, scores: array1D):
        _pop, _score = zip(
            *[
                local_optimizer(population[i], scores[i]) for i in range(scores.size)
            ]
        )
        return np.array(_pop), np.array(_score)


    #def Create_population(func, start_generation, expected_size, #variable_boundaries):
    #    
    #    if not (start_generation['variables'] is None):
    #        pop = start_generation['variables']
    #        scores = start_generation['scores']
    #        if scores is None:
    #            scores = np.array([func(pop[i, :]) for i in range(pop.shape[0])])
    #        return pop, scores


    def Result(population: array2D, scores: array1D):
        if local_optimization_step == 'before_select':
            pop, s = Local_opt(population, scores)
            return Select_best(pop, s)

        if local_optimization_step == 'after_select':
            pop, s = Select_best(population, scores)
            return Local_opt(pop, s)

        #if local_optimization_step == 'never':
        return Select_best(population, scores)
    

    return select_best_of, Result












from typing import Optional, Any, Tuple

from pathlib import Path

import random
import numpy as np

from .aliases import array1D, array2D


def fast_min(a, b):
    '''
    1.5 times faster than row min(a, b)
    '''
    return a if a < b else b


def fast_max(a, b):
    return a if a > b else b


def mkdir(folder: str):
    Path(folder).mkdir(parents=True, exist_ok=True)


def can_be_prob(value: float) -> bool:
    return 0 <= value <= 1


def is_current_gen_number(number: Optional[int]):
    return (number is None) or (type(number) == int and number > 0)


def is_numpy(arg: Any):
    return type(arg) == np.ndarray


def split_matrix(mat: array2D) -> Tuple[array2D, array1D]:
    """
    splits wide pop matrix to variables and scores
    """
    return mat[:, :-1], mat[:, -1]


def union_to_matrix(variables_2D: array2D, scores_1D: array1D) -> array2D:
    """
    union variables and scores to wide pop matrix
    """
    return np.hstack((variables_2D, scores_1D[:, np.newaxis]))


def random_indexes_pair(seq_len: int) -> Tuple[int, int]:
    """works 3 times faster than `random.sample(range(seq_len), 2)`"""
    a = random.randrange(seq_len)
    b = random.randrange(seq_len)
    if a == b:
        while a == b:
            b = random.randrange(seq_len)
    return a, b

















import sys
from typing import List, Tuple, Dict, Sequence, Optional, Any, Callable, Union, TypeVar, Literal

if sys.version_info.minor < 10:
    from typing_extensions import TypeAlias
else:
    from typing import TypeAlias

Number: TypeAlias = Union[int, float]

import numpy as np

array1D: TypeAlias = np.ndarray
array2D: TypeAlias = np.ndarray





from typing import Dict, Any, List, Optional, Union, Callable, Tuple, Literal


from dataclasses import dataclass
import warnings

import numpy as np

from .aliases import array1D, array2D, TypeAlias

from .crossovers import Crossover, CrossoverFunc
from .mutations import Mutations, MutationIntFunc, MutationFloatFunc
from .selections import Selection, SelectionFunc

from .utils import can_be_prob, union_to_matrix


class DictLikeGetSet:
    def __getitem__(self, item):
        return getattr(self, item)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def get(self, item):
        return getattr(self, item)



_algorithm_params_slots = {
    'max_num_iteration',
    'max_iteration_without_improv',
    'population_size',
    'mutation_probability',
    'mutation_discrete_probability',
    'elit_ratio',
    'crossover_probability',
    'parents_portion',
    'crossover_type',
    'mutation_type',
    'mutation_discrete_type',
    'selection_type'
}


@dataclass
class AlgorithmParams(DictLikeGetSet):

    max_num_iteration: Optional[int] = None
    max_iteration_without_improv: Optional[int] = None

    population_size: int = 100

    mutation_probability: float = 0.1
    mutation_discrete_probability: Optional[float] = None

    #  deprecated
    crossover_probability: Optional[float] = None

    elit_ratio: float = 0.04
    parents_portion: float = 0.3

    crossover_type: Union[str, CrossoverFunc] = 'uniform'
    mutation_type: Union[str, MutationFloatFunc] = 'uniform_by_center'
    mutation_discrete_type: Union[str, MutationIntFunc] = 'uniform_discrete'
    selection_type: Union[str, SelectionFunc] = 'roulette'

    __annotations__ = {
        'max_num_iteration': Optional[int],
        'max_iteration_without_improv': Optional[int],
        'population_size': int,
        'mutation_probability': float,
        'mutation_discrete_probability': Optional[float],
        'crossover_probability': Optional[float],
        'elit_ratio': float,
        'parents_portion': float,
        'crossover_type': Union[str, CrossoverFunc],
        'mutation_type': Union[str, MutationFloatFunc],
        'mutation_discrete_type': Union[str, MutationIntFunc],
        'selection_type': Union[str, SelectionFunc]
    }

    def check_if_valid(self) -> None:

        assert int(self.population_size) > 0, f"population size must be integer and >0, not {self.population_size}"
        assert (can_be_prob(self.parents_portion)), "parents_portion must be in range [0,1]"
        assert (can_be_prob(self.mutation_probability)), "mutation_probability must be in range [0,1]"
        assert (can_be_prob(self.elit_ratio)), "elit_ratio must be in range [0,1]"

        if self.max_iteration_without_improv is not None and self.max_iteration_without_improv < 1:
            warnings.warn(f"max_iteration_without_improv is {self.max_iteration_without_improv} but must be None or int > 0")
            self.max_iteration_without_improv = None

    def get_CMS_funcs(self) -> Tuple[
        CrossoverFunc,
        MutationFloatFunc,
        MutationIntFunc,
        SelectionFunc
    ]:
        """
        returns gotten crossover, mutation, discrete mutation, selection
        as necessary functions
        """

        result: List[Callable] = []
        for name, value, dct in (
            ('crossover', self.crossover_type, Crossover.crossovers_dict()),
            ('mutation', self.mutation_type, Mutations.mutations_dict()),
            ('mutation_discrete', self.mutation_discrete_type, Mutations.mutations_discrete_dict()),
            ('selection', self.selection_type, Selection.selections_dict())
        ):
            if type(value) == str:
                if value not in dct:
                    raise ValueError(
                        f"unknown name of {name}: '{value}', must be from {tuple(dct.keys())} or a custom function"
                    )
                result.append(dct[value])
            else:
                assert callable(value), f"{name} must be string or callable"
                result.append(value)

        return tuple(result)

    @staticmethod
    def from_dict(dct: Dict[str, Any]):

        result = AlgorithmParams()

        for name, value in dct.items():
            if name not in _algorithm_params_slots:
                raise AttributeError(f"name '{name}' does not exists in AlgorithmParams fields")

            setattr(result, name, value)
        return result


GenerationConvertible: TypeAlias = Union[
    'Generation',
    str,
    Dict[Literal['population', 'scores'], Union[array2D, array1D]],
    array2D,
    Tuple[
        Optional[array2D],
        Optional[array1D]
    ]
]


@dataclass
class Generation(DictLikeGetSet):
    variables: Optional[array2D] = None
    scores: Optional[array1D] = None

    __annotations__ = {
        'variables': Optional[array2D],
        'scores': Optional[array1D]
    }

    def __check_dims(self) -> None:
        if self.variables is not None:
            assert len(self.variables.shape) == 2, f"'variables' must be matrix with shape (objects, dimensions), not {self.variables.shape}"
            if self.scores is not None:
                assert len(self.scores.shape) == 1, f"'scores' must be 1D-array, not with shape {self.scores.shape}"
                assert self.variables.shape[0] == self.scores.size, f"count of objects ({self.variables.shape[0]}) must be equal to count of scores ({self.scores.size})"

    @property
    def size(self) -> int:
        return self.scores.size

    @property
    def dim_size(self) -> int:
        return self.variables.shape[1]

    def as_wide_matrix(self) -> array2D:
        # should not be used in main code -- was needed for old versions
        return union_to_matrix(self.variables, self.scores)

    def save(self, path: str):
        np.savez(path, population=self.variables, scores=self.scores)

    @staticmethod
    def load(path: str):
        try:
            st = np.load(path)
        except Exception as err:
            raise Exception(
                f"if generation object is a string, it must be path to npz file with needed content, but raised exception {repr(err)}"
            )

        assert 'population' in st and 'scores' in st, "saved generation object must contain 'population' and 'scores' fields"

        return Generation(variables=st['population'], scores=st['scores'])

    @staticmethod
    def from_object(
        dim: int,
        object: GenerationConvertible
    ):

        obj_type = type(object)

        if obj_type == str:

            generation = Generation.load(object)

        elif obj_type == np.ndarray:

            assert len(object.shape) == 2 and (object.shape[1] == dim or object.shape[1] == dim + 1), f"if start_generation is numpy array, it must be with shape (samples, dim) or (samples, dim+1), not {object.shape}"

            generation = Generation(object, None) if object.shape[1] == dim else Generation.from_pop_matrix(object)

        elif obj_type == tuple:

            assert len(object) == 2, f"if start_generation is tuple, it must be tuple with 2 components, not {len(object)}"

            variables, scores = object

            assert ( (variables is None or scores is None) or (
                            variables.shape[0] == scores.size)), "start_generation object must contain variables and scores components which are None or 2D- and 1D-arrays with same shape"

            generation = Generation(variables=variables, scores=scores)

        elif obj_type == dict:
            assert (
                ('variables' in object and 'scores' in object) and
                (object['variables'] is None or object['scores'] is None) or
                (object['variables'].shape[0] == object['scores'].size)
            ), "start_generation object must contain 'variables' and 'scores' keys which are None or 2D- and 1D-arrays with same shape"

            generation = Generation(variables=object['variables'], scores=object['scores'])

        elif obj_type == Generation:
            generation = Generation(variables=object['variables'], scores=object['scores'])
        else:
            raise TypeError(f"invalid type of generation! Must be in (Union[str, Dict[str, np.ndarray], Generation, np.ndarray, Tuple[Optional[np.ndarray], Optional[np.ndarray]]]), not {obj_type}")

        generation.__check_dims()

        if generation.variables is not None:
            assert generation.dim_size == dim, f"generation dimension size {generation.dim_size} does not equal to target size {dim}"

        return generation

    @staticmethod
    def from_pop_matrix(pop: array2D):
        warnings.warn("depricated! pop matrix style will be removed at version 7, use samples and scores separetly")
        return Generation(
            variables=pop[:, :-1],
            scores=pop[:, -1]
        )


@dataclass
class GAResult(DictLikeGetSet):

    last_generation: Generation

    __annotations__ = {
        'last_generation': Generation
    }

    @property
    def variable(self) -> array1D:
        return self.last_generation.variables[0]

    @property
    def score(self) -> float:
        return self.last_generation.scores[0]

    @property
    def function(self):
        warnings.warn(f"'function' field is deprecated, will be removed in version 7, use 'score' to get best population score")
        return self.score


@dataclass
class MiddleCallbackData(DictLikeGetSet):
    """
    data object using with middle callbacks
    """

    reason_to_stop: Optional[str]

    last_generation: Generation

    current_generation: int
    report_list: List[float]

    mutation_prob: float
    mutation_discrete_prob: float

    mutation: MutationFloatFunc
    mutation_discrete: MutationIntFunc
    crossover: CrossoverFunc
    selection: SelectionFunc

    current_stagnation: int
    max_stagnation: int

    parents_portion: float
    elit_ratio: float

    set_function: Callable[[array2D], array1D]
    
    __annotations__ = {
        'reason_to_stop': Optional[str],
        'last_generation': Generation,
        'current_generation': int,
        'report_list': List[float],
        'mutation_prob': float,
        'mutation_discrete_prob': float,
        'mutation': MutationFloatFunc,
        'mutation_discrete': MutationIntFunc,
        'crossover': CrossoverFunc,
        'selection': SelectionFunc,
        'current_stagnation': int,
        'max_stagnation': int,
        'parents_portion': float,
        'elit_ratio': float,
        'set_function': Callable[[array2D], array1D]
    }


















from typing import Callable, Dict, Union

import random
import numpy as np


from .utils import fast_min, fast_max

from .aliases import TypeAlias

MutationFloatFunc: TypeAlias = Callable[[float, float, float], float]
MutationIntFunc: TypeAlias = Callable[[int, int, int], int]
MutationFunc: TypeAlias = Union[MutationIntFunc, MutationFloatFunc]


class Mutations:

    @staticmethod
    def mutations_dict() -> Dict[str, MutationFloatFunc]:
        return {
            'uniform_by_x': Mutations.uniform_by_x(),
            'uniform_by_center': Mutations.uniform_by_center(),
            'gauss_by_center': Mutations.gauss_by_center(),
            'gauss_by_x': Mutations.gauss_by_x(),
        }

    @staticmethod
    def mutations_discrete_dict() -> Dict[str, MutationIntFunc]:
        return {
            'uniform_discrete': Mutations.uniform_discrete()
        }


    @staticmethod
    def uniform_by_x() -> MutationFloatFunc:
        
        def func(x: float, left: float, right: float):
            alp: float = fast_min(x - left, right - x)
            return random.uniform(x - alp, x + alp)
        return func

    @staticmethod
    def uniform_by_center() -> MutationFloatFunc:
        
        def func(x: float, left: float, right: float):
            return random.uniform(left, right)
        
        return func

    @staticmethod
    def gauss_by_x(sd: float = 0.3) -> MutationFloatFunc:
        """
        gauss mutation with x as center and sd*length_of_zone as std
        """
        def func(x: float, left: float, right: float):
            std: float = sd * (right - left)
            return fast_max(
                left, 
                fast_min(right, np.random.normal(loc=x, scale=std))
            )
        
        return func

    @staticmethod
    def gauss_by_center(sd: float = 0.3) -> MutationFloatFunc:
        """
        gauss mutation with (left+right)/2 as center and sd*length_of_zone as std
        """
        def func(x: float, left: float, right: float):
            std: float = sd * (right - left)
            return fast_max(
                left, 
                fast_min(right, np.random.normal(loc=(left + right) * 0.5, scale=std))
            )
        
        return func

    @staticmethod
    def uniform_discrete() -> MutationIntFunc:
        def func(x: int, left: int, right: int) -> int:
            return random.randint(left, right)
        return func





from typing import List, Optional, Callable, Tuple, Sequence

import os
import random

import numpy as np

from OppOpPopInit import OppositionOperators, SampleInitializers

from .aliases import TypeAlias, array1D, array2D

from .classes import MiddleCallbackData, Generation

from .utils import mkdir, union_to_matrix, fast_max

from .crossovers import CrossoverFunc
from .selections import SelectionFunc
from .mutations import MutationFunc


CallbackFunc: TypeAlias = Callable[[int, List[float], array2D, array1D], None]
MiddleCallbackActionFunc: TypeAlias = Callable[[MiddleCallbackData], MiddleCallbackData]
MiddleCallbackConditionFunc: TypeAlias = Callable[[MiddleCallbackData], bool]
MiddleCallbackFunc: TypeAlias = Callable[[MiddleCallbackData], Tuple[MiddleCallbackData, bool]]


class Callbacks:
    
    @staticmethod
    def NoneCallback():
        return lambda generation_number, report_list, last_population, last_scores: None

    @staticmethod
    def SavePopulation(folder: str, save_gen_step: int = 50, file_prefix: str = 'population') -> CallbackFunc:
        
        mkdir(folder)

        def func(generation_number: int, report_list: List[float], last_population: array2D, last_scores: array1D):
    
            if generation_number % save_gen_step != 0:
                return

            np.savez(os.path.join(
                folder,
                f"{file_prefix}_{generation_number}.npz"),
                population=last_population,
                scores=last_scores
            )
        
        return func
    
    @staticmethod
    def PlotOptimizationProcess(
        folder: str,
        save_gen_step: int = 50,
        show: bool = False,
        main_color: str = 'green',
        file_prefix: str = 'report'
    ) -> CallbackFunc:
        import matplotlib.pyplot as plt
        from matplotlib.ticker import MaxNLocator
        
        mkdir(folder)

        def func(generation_number: int, report_list: List[float], last_population: array2D, last_scores: array1D):

            if generation_number % save_gen_step != 0:
                return

            # if len(report_list) == 0:
            #     sys.stdout.write("No results to plot!\n")
            #     return
            
            ax = plt.axes()
            ax.xaxis.set_major_locator(MaxNLocator(integer=True))
            
            plt.plot(
                np.arange(1, 1 + len(report_list)),
                report_list,
                color=main_color,
                label='best of generation',
                linewidth=2
            )
            
            plt.xlabel('Generation')
            plt.ylabel('Minimized function')
            plt.title('GA optimization process')
            plt.legend()
            
            plt.savefig(os.path.join(folder, f"{file_prefix}_{generation_number}.png"), dpi=200)

            if show:
                plt.show()
            else:
                plt.close()
        
        return func


class Actions:

    @staticmethod
    def Stop(reason_name: str = 'stopped by Stop callback') -> MiddleCallbackActionFunc:
        
        def func(data: MiddleCallbackData):
            data.reason_to_stop = reason_name
            return data
        return func

    @staticmethod
    def ReduceMutationProb(reduce_coef: float = 0.9) -> MiddleCallbackActionFunc:
        
        def func(data: MiddleCallbackData):
            data.mutation_prob *= reduce_coef
            return data
        
        return func


    #def DualStrategyStep():
    #    pass

    #def SetFunction():
    #    pass


    @staticmethod
    def ChangeRandomCrossover(
        available_crossovers: Sequence[CrossoverFunc]
    ) -> MiddleCallbackActionFunc:

        def func(data: MiddleCallbackData):
            data.crossover = random.choice(available_crossovers)
            return data

        return func
    
    @staticmethod
    def ChangeRandomSelection(available_selections: Sequence[SelectionFunc]) -> MiddleCallbackActionFunc:

        def func(data: MiddleCallbackData):
            data.selection = random.choice(available_selections)
            return data

        return func

    @staticmethod
    def ChangeRandomMutation(available_mutations: Sequence[MutationFunc]) -> MiddleCallbackActionFunc:

        def func(data):
            data.mutation = random.choice(available_mutations)
            return data

        return func
    

    @staticmethod
    def RemoveDuplicates(
        oppositor: Optional[Callable[[array1D], array1D]] = None,
        creator: Optional[Callable[[], array1D]] = None,
        converter: Optional[Callable[[array1D], array1D]] = None
    ) -> MiddleCallbackActionFunc:
        """
        Removes duplicates from population

        Parameters
        ----------
        oppositor : oppositor from OppOpPopInit, optional
            oppositor for applying after duplicates removing. By default -- using just random initializer from creator. The default is None.
        creator : the function creates population samples, optional
            the function creates population samples if oppositor is None. The default is None.
        converter : func, optional
            function converts population samples in new format to compare (if needed). The default is None.

        """

        if creator is None and oppositor is None:
            raise Exception("No functions to fill population! creator or oppositors must be not None")

        if converter is None:
            def without_dup(pop: array2D, scores: array1D) -> Tuple[array2D, int]:
                """returns population without dups"""
                _, index_of_dups = np.unique(pop, axis=0, return_index=True)
                return union_to_matrix(pop[index_of_dups], scores[index_of_dups]), pop.shape[0] - index_of_dups.size
        else:
             def without_dup(pop: array2D, scores: array1D) -> Tuple[array2D, int]:
                """returns population without dups"""
                _, index_of_dups = np.unique(
                    np.array([converter(pop[i]) for i in range(pop.shape[0])]), axis=0, return_index=True
                )
                return union_to_matrix(pop[index_of_dups], scores[index_of_dups]), pop.shape[0] - index_of_dups.size


        if oppositor is None:
            def remover(pop: array2D, scores: array1D, set_function: Callable[[array2D], array1D]) -> array2D:

                pp, count_to_create = without_dup(pop, scores)  # pop without dups
                pp2 = np.empty((count_to_create, pp.shape[1])) 
                pp2[:, :-1] = SampleInitializers.CreateSamples(creator, count_to_create)  # new pop elements
                pp2[:, -1] = set_function(pp2[:, :-1])  # new elements values

                new_pop = np.vstack((pp, pp2))

                return new_pop[np.argsort(new_pop[:, -1])]  # new pop

        else:  # using oppositors
            def remover(pop: array2D, scores: array1D, set_function: Callable[[array2D], array1D]) -> array2D:

                pp, count_to_create = without_dup(pop, scores)  # pop without dups

                if count_to_create > pp.shape[0]:
                    raise Exception("Too many duplicates, cannot oppose")
                
                if count_to_create == 0:
                    return pp[np.argsort(pp[:, -1])]
                
                pp2 = np.empty((count_to_create, pp.shape[1])) 
                # oppose count_to_create worse elements
                pp2[:, :-1] = OppositionOperators.Reflect(pp[-count_to_create:, :-1], oppositor)  # new pop elements
                pp2[:, -1] = set_function(pp2[:, :-1])  # new elements values
                    
                new_pop = np.vstack((pp, pp2))
                    
                return new_pop[np.argsort(new_pop[:, -1])]  # new pop

        def func(data: MiddleCallbackData):
            new_pop = remover(
                data.last_generation.variables,
                data.last_generation.scores,
                data.set_function
            )

            data.last_generation = Generation.from_pop_matrix(new_pop)

            return data

        
        return func
    
    @staticmethod
    def CopyBest(by_indexes: Sequence[int]) -> MiddleCallbackActionFunc:
        """
        Copies best population object values (from dimensions in by_indexes) to all population
        """

        if type(by_indexes) != np.ndarray:
            by_indexes = np.array(by_indexes)

        def func(data: MiddleCallbackData):

            pop = data.last_generation.variables
            scores = data.last_generation.scores

            pop[:, by_indexes] = pop[np.argmin(scores), by_indexes]

            data.last_generation = Generation(pop, data.set_function(pop))

            return data
        return func

    @staticmethod
    def PlotPopulationScores(
        title_pattern: Callable[[MiddleCallbackData], str] = lambda data: f"Generation {data.current_generation}",
        save_as_name_pattern: Optional[Callable[[MiddleCallbackData], str]] = None
    ) -> MiddleCallbackActionFunc:
        """
        plots population scores
        needs 2 functions like data->str for title and file name
        """
        from .plotting_tools import plot_pop_scores

        use_save_as = (lambda data: None) if save_as_name_pattern is None else save_as_name_pattern

        def local_plot_callback(data: MiddleCallbackData):

            plot_pop_scores(
                data.last_generation.scores,
                title = title_pattern(data),
                save_as = use_save_as(data)
            )

            return data

        return local_plot_callback


class ActionConditions:
    
    @staticmethod
    def EachGen(generation_step: int = 10) -> MiddleCallbackConditionFunc:

        if generation_step < 1 or type(generation_step) != int:
            raise Exception(f"Invalid generation step {generation_step}! Should be int and >=1")
        
        if generation_step == 1:
            return ActionConditions.Always()

        def func(data: MiddleCallbackData):
            return data.current_generation % generation_step == 0 and data.current_generation > 0

        return func

    @staticmethod
    def Always() -> MiddleCallbackConditionFunc:
        """
        makes action each generation
        """
        return lambda data: True

    @staticmethod
    def AfterStagnation(stagnation_generations: int = 50) -> MiddleCallbackConditionFunc:

        def func(data: MiddleCallbackData):
            return data.current_stagnation % stagnation_generations == 0 and data.current_stagnation > 0
        return func


    @staticmethod
    def Several(conditions: Sequence[MiddleCallbackConditionFunc]) -> MiddleCallbackConditionFunc:
        """
        returns function which checks all conditions from conditions
        """

        def func(data: MiddleCallbackData):
            return all(cond(data) for cond in conditions)
        
        return func


class MiddleCallbacks:

    @staticmethod
    def UniversalCallback(
        action: MiddleCallbackActionFunc,
        condition: MiddleCallbackConditionFunc,
        set_data_after_callback: bool = True
    ) -> MiddleCallbackFunc:
        
        def func(data: MiddleCallbackData):

            cond = condition(data)
            if cond:
                data = action(data)

            return data, (cond if set_data_after_callback else False)
        
        return func

    @staticmethod
    def ReduceMutationGen(
        reduce_coef: float = 0.9,
        min_mutation: float = 0.005,
        reduce_each_generation: int = 50,
        reload_each_generation: int = 500
    ) -> MiddleCallbackFunc:

        start_mutation = None

        def func(data: MiddleCallbackData):
            nonlocal start_mutation
            
            gen = data.current_generation
            mut = data.mutation_prob

            if start_mutation is None:
                start_mutation = mut

            c1 = gen % reduce_each_generation == 0
            c2 = gen % reload_each_generation == 0

            if c2:
                mut = start_mutation
            elif c1:
                mut *= reduce_coef
                mut = fast_max(mut, min_mutation)

            data.mutation_prob = mut

            return data, (c1 or c2)

        return func

    #def ReduceMutationStagnation(reduce = 0.5, stagnation_gens = 50):
    #    pass 

    @staticmethod
    def GeneDiversityStats(step_generations_for_plotting: int = 10) -> MiddleCallbackFunc:

        if step_generations_for_plotting < 1:
            raise Exception(f"Wrong step = {step_generations_for_plotting}, should be int and > 0!!")

        import matplotlib.pyplot as plt

        div = []
        count = []
        most = []


        def func(data: MiddleCallbackData):

            nonlocal div, count, most

            dt = data.last_generation.variables

            uniq, counts = np.unique(dt, return_counts=True, axis=0)
            # raise Exception()
            count.append(counts.size)
            most.append(counts.max())

            gene_diversity = 0
            for index in range(dt.shape[0]-1):
                gene_diversity += np.count_nonzero(dt[index, :] != dt[index:, :]) / (dt.shape[0] - index)
            div.append(gene_diversity/dt.shape[1])

            if data.current_generation % step_generations_for_plotting == 0:

                fig, axes = plt.subplots(3, 1)

                (ax1, ax2, ax3) = axes

                ax1.plot(count)
                #axs[0, 0].set_title('Axis [0, 0]')

                ax2.plot(most, 'tab:orange')
                #axs[0, 1].set_title('Axis [0, 1]')

                ax3.plot(div, 'tab:green')
                #axs[1, 0].set_title('Axis [1, 0]')

                ylabs = [
                    'Count of unique objects',
                    'Count of most popular object',
                    'Simple gene diversity'
                ]

                for i, ax in enumerate(axes):
                    ax.set(xlabel='Generation number')
                    ax.set_title(ylabs[i])

                # Hide x labels and tick labels for top plots and y ticks for right plots.
                for ax in axes:
                    ax.label_outer()

                fig.suptitle(f'Diversity report (pop size = {dt.shape[0]})')    
                fig.tight_layout()
                plt.show()

            return data, False

        return func






